"""OVK learning, unit tests.

The :mod:`sklearn.tests.test_simplex_coding` tests the simplex coding
estimator.
"""
from numpy import zeros, array

import operalib as ovk

from sklearn.utils.testing import assert_array_almost_equal, assert_array_equal


def test_simplexcoding():
    """Test SimplexCoding."""
    # one-class case
    inp = ["pos", "pos", "pos", "pos"]
    scode = ovk.preprocessing.SimplexCoding()
    expected = zeros((4, 1))
    got = scode.fit_transform(inp)
    assert_array_equal(scode.binarizer_.classes_, ["pos"])
    assert_array_equal(expected, got)
    assert_array_equal(scode.inverse_transform(got), inp)

    # two-class case
    inp = ["neg", "pos", "pos", "neg"]
    expected = array([[0], [1], [1], [0]])
    got = scode.fit_transform(inp)
    assert_array_equal(scode.binarizer_.classes_, ["neg", "pos"])
    assert_array_equal(expected, got)
    assert_array_equal(scode.inverse_transform(got), inp)

    # multi-class case
    inp = ["spam", "ham", "eggs", "ham", "0"]
    expected = array([[-1./3., -0.471, -0.816],
                      [-1./3., -0.471, 0.816],
                      [-1./3., 0.943, 0.],
                      [-1./3., -0.471, 0.816],
                      [1., 0., 0.]])
    got = scode.fit_transform(inp)
    assert_array_equal(scode.binarizer_.classes_, ['0', 'eggs', 'ham', 'spam'])
    assert_array_almost_equal(expected, got, decimal=1e-2)
    assert_array_equal(scode.inverse_transform(got), inp)



"""
===================================================
Vector-field Learning with structured output kernel
===================================================

An example to illustrate structured learning with operator-valued kernels.

We compare Operator-valued kernel (OVK) with scikit-learn multi-output ridge
regression on a semi-supervised dataset.
"""
# Author: Romain Brault <ro.brault@gmail.com>
# License: MIT

# -*- coding: utf-8 -*-

import operalib as ovk

import numpy as np
from numpy.random import RandomState

import matplotlib.pyplot as plt

from sklearn.model_selection import train_test_split
from sklearn.kernel_ridge import KernelRidge


def main():
    """Example of vector-field learning."""

    # Fix a seed
    random_state = RandomState(0)

    # Generate data
    inputs, targets = ovk.toy_data_curl_free_field(n_samples=2000)
    inputs_mesh = ovk.array2mesh(inputs)

    (inputs_train, inputs_test,
     targets_train, targets_test) = train_test_split(inputs, targets,
                                                     train_size=(inputs
                                                                 .shape[0] -
                                                                 40 ** 2),
                                                     random_state=random_state)

    # Add unlabelled data
    targets_train = ovk.datasets.awful(targets_train, .85, .1, .5,
                                       random_state=random_state)
    sup_mask = ~np.any(np.isnan(targets_train), axis=1)
    weaksup_mask = ~np.all(np.isnan(targets_train), axis=1)

    scores = {'CF weak semisup': 0, 'CF weak': 0, 'CF': 0, 'Indep': 0}
    regressor = {'CF weak semisup':
                 ovk.OVKRidge(ovkernel=ovk.RBFCurlFreeKernel(gamma=2.),
                              lbda=0.0001, gamma_m=1., lbda_m=0.0001),
                 'CF weak':
                 ovk.OVKRidge(ovkernel=ovk.RBFCurlFreeKernel(gamma=2.),
                              lbda=0.0001),
                 'CF':
                 ovk.OVKRidge(ovkernel=ovk.RBFCurlFreeKernel(gamma=2.),
                              lbda=0.0001),
                 'Indep':
                 KernelRidge(kernel='rbf', alpha=0.0001, gamma=.5)}
    targets_mesh = {'CF weak semisup': (None, None),
                    'CF weak': (None, None),
                    'CF': (None, None),
                    'Indep': (None, None)}

    # Learning with curl-free semisup + weak
    regressor['CF weak semisup'].fit(inputs_train, targets_train)
    scores['CF weak semisup'] = (regressor['CF weak semisup']
                                 .score(inputs_test, targets_test))
    print('R2 curl-free semisup + weak ridge: ', scores['CF weak semisup'])
    targets_mesh['CF weak semisup'] = ovk.array2mesh(
        regressor['CF weak semisup'].predict(inputs))

    # Learning with curl-free weak
    regressor['CF weak'].fit(inputs_train[weaksup_mask, :],
                             targets_train[weaksup_mask, :])
    scores['CF weak'] = regressor['CF weak'].score(inputs_test, targets_test)
    print('R2 curl-free weak ridge: ', scores['CF weak'])
    targets_mesh['CF weak'] = ovk.array2mesh(
        regressor['CF weak'].predict(inputs))

    # Learning with curl-free
    regressor['CF'].fit(inputs_train[sup_mask, :], targets_train[sup_mask, :])
    scores['CF'] = regressor['CF'].score(inputs_test, targets_test)
    print('R2 curl-free ridge: ', scores['CF'])
    targets_mesh['CF'] = ovk.array2mesh(regressor['CF'].predict(inputs))

    # Learning with sklearn ridge
    regressor['Indep'].fit(inputs_train[sup_mask, :],
                           targets_train[sup_mask, :])
    scores['Indep'] = regressor['Indep'].score(inputs_test, targets_test)
    print('R2 independant ridge: ', scores['Indep'])
    targets_mesh['Indep'] = ovk.array2mesh(regressor['Indep'].predict(inputs))

    # Plotting
    # pylint: disable=E1101
    fig, axarr = plt.subplots(2, 2, sharex=True, sharey=True, figsize=(14, 14))
    axarr[0, 0].streamplot(inputs_mesh[0], inputs_mesh[1],
                           targets_mesh['CF weak semisup'][0],
                           targets_mesh['CF weak semisup'][1],
                           color=np.sqrt(
                               targets_mesh['CF weak semisup'][0]**2 +
                               targets_mesh['CF weak semisup'][1]**2),
                           linewidth=.5, cmap=plt.cm.jet, density=2,
                           arrowstyle=u'->')
    axarr[0, 1].streamplot(inputs_mesh[0], inputs_mesh[1],
                           targets_mesh['CF weak'][0],
                           targets_mesh['CF weak'][1],
                           color=np.sqrt(targets_mesh['CF weak'][0]**2 +
                                         targets_mesh['CF weak'][1]**2),
                           linewidth=.5, cmap=plt.cm.jet, density=2,
                           arrowstyle=u'->')
    axarr[1, 0].streamplot(inputs_mesh[0], inputs_mesh[1],
                           targets_mesh['CF'][0], targets_mesh['CF'][1],
                           color=np.sqrt(targets_mesh['CF'][0]**2 +
                                         targets_mesh['CF'][1]**2),
                           linewidth=.5, cmap=plt.cm.jet, density=2,
                           arrowstyle=u'->')
    axarr[1, 1].streamplot(inputs_mesh[0], inputs_mesh[1],
                           targets_mesh['Indep'][0], targets_mesh['Indep'][1],
                           color=np.sqrt(targets_mesh['Indep'][0]**2 +
                                         targets_mesh['Indep'][1]**2),
                           linewidth=.5, cmap=plt.cm.jet, density=2,
                           arrowstyle=u'->')
    axarr[0, 0].set_ylim([-1, 1])
    axarr[0, 0].set_xlim([-1, 1])
    axarr[0, 0].set_title('Curl-free semisup + weak ridge: ' +
                          str(scores['CF weak semisup']))
    axarr[0, 1].set_ylim([-1, 1])
    axarr[0, 1].set_xlim([-1, 1])
    axarr[0, 1].set_title('Curl-free weak ridge, R2: ' +
                          str(scores['CF weak']))
    axarr[1, 0].set_ylim([-1, 1])
    axarr[1, 0].set_xlim([-1, 1])
    axarr[1, 0].set_title('Curl-free ridge, R2: ' +
                          str(scores['CF']))
    axarr[1, 1].set_ylim([-1, 1])
    axarr[1, 1].set_xlim([-1, 1])
    axarr[1, 1].set_title('Independant ridge, R2: ' +
                          str(scores['Indep']))
    fig.suptitle('Vectorfield learning')
    plt.show()

if __name__ == '__main__':
    main()



"""OVK learning.

The :mod:`operalib` module includes algorithms to learn with Operator-valued
kernels.
"""
# Author: Romain Brault <romain.brault@telecom-paristech.fr> with help from
#         the scikit-learn community.
# License: MIT

import sys
import re
import warnings

# Make sure that DeprecationWarning within this package always gets printed
warnings.filterwarnings('always', category=DeprecationWarning,
                        module=r'^{0}\.'.format(re.escape(__name__)))

# PEP0440 compatible formatted version, see:
# https://www.python.org/dev/peps/pep-0440/
#
# Generic release markers:
#   X.Y
#   X.Y.Z   # For bugfix releases
#
# Admissible pre-release markers:
#   X.YaN   # Alpha release
#   X.YbN   # Beta release
#   X.YrcN  # Release Candidate
#   X.Y     # Final release
#
# Dev branch marker is: 'X.Y.dev' or 'X.Y.devN' where N is an integer.
# 'X.Y.dev0' is the canonical version of 'X.Y.dev'
#

__author__ = ['Romain Brault']
__copyright__ = 'Copyright 2015, Operalib'
__credits__ = ['Romain Brault', 'Florence D\'Alche Buc',
               'Markus Heinonen', 'Tristan Tchilinguirian',
               'Alexandre Gramfort']
__license__ = 'MIT'
__version__ = '0.2b26'
__maintainer__ = ['Romain Brault']
__email__ = ['romain.brault@telecom-paristech.fr']
__status__ = 'Beta'

try:
    # This variable is injected in the __builtins__ by the build
    # process. It used to enable importing subpackages of sklearn when
    # the binaries are not built
    __OPERALIB_SETUP__
except NameError:
    __OPERALIB_SETUP__ = False

if __OPERALIB_SETUP__:
    sys.stderr.write('Partial import of operalib during the build process.\n')
    # We are not importing the rest of Operalib during the build
    # process, as it may not be compiled yet
else:
    from .kernels import (DecomposableKernel, RBFCurlFreeKernel,
                          RBFDivFreeKernel, DotProductKernel)
    from .risk import OVKRidgeRisk, ORFFRidgeRisk, ORFFLSLoss, ORFFHingeLoss
    from .learningrate import Constant, InvScaling
    from .ridge import OVKRidge
    from .onorma import ONORMA
    from .orff import ORFFRidge
    from .quantile import Quantile
    from .metrics import first_periodic_kernel
    from .signal import get_period
    from .datasets.vectorfield import (toy_data_curl_free_field,
                                       toy_data_curl_free_mesh,
                                       toy_data_div_free_field,
                                       toy_data_div_free_mesh,
                                       array2mesh, mesh2array)
    from .datasets.quantile import toy_data_quantile
    from .preprocessing.simplex import SimplexCoding

    __all__ = ['DecomposableKernel', 'RBFCurlFreeKernel', 'RBFDivFreeKernel',
               'DotProductKernel',
               'OVKRidgeRisk', 'ORFFRidgeRisk', 'ORFFLSLoss', 'ORFFHingeLoss',
               'Constant', 'InvScaling',
               'first_periodic_kernel',
               'get_period',
               'OVKRidge', 'ORFFRidge', 'Quantile', 'ONORMA',
               'toy_data_curl_free_field', 'toy_data_curl_free_mesh',
               'toy_data_div_free_field', 'toy_data_div_free_mesh',
               'toy_data_quantile',
               'array2mesh', 'mesh2array',
               'SimplexCoding']



"""
======================================================
Joint quantile regression with operator-valued kernels
======================================================

An example to illustrate joint quantile regression with operator-valued
kernels.

We compare quantile regression estimation with and without non-crossing
constraints.
"""

# Author: Maxime Sangnier <maxime.sangnier@gmail.com>
# License: MIT

# -*- coding: utf-8 -*-
import time

import numpy as np
import matplotlib.pyplot as plt

from operalib import Quantile, toy_data_quantile


def main():
    """Example of multiple quantile regression."""

    print("Creating dataset...")
    probs = np.linspace(0.1, 0.9, 5)  # Quantile levels of interest
    x_train, y_train, _ = toy_data_quantile(100, random_state=0)
    x_test, y_test, z_test = toy_data_quantile(1000, probs=probs,
                                               random_state=1)

    print("Fitting...")
    methods = {'Default': Quantile(),
                'Joint':
               Quantile(probs=probs, kernel='DGauss', lbda=1e-2, gamma=8.,
                        gamma_quantile=1e-2),
               'Independent': Quantile(probs=probs, kernel='DGauss',
                                       lbda=1e-2, gamma=8.,
                                       gamma_quantile=np.inf),
               'Non-crossing': Quantile(probs=probs, kernel='DGauss',
                                        lbda=1e-2, gamma=8.,
                                        gamma_quantile=np.inf, nc_const=True)}
    # Fit on training data
    for name, reg in sorted(methods.items()):
        print(name)
        start = time.time()
        reg.fit(x_train, y_train)
        print(name + ' leaning time: ', time.time() - start)
        print(name + ' score ', reg.score(x_test, y_test))

    # Plot the estimated conditional quantiles
    plt.figure(figsize=(12, 7))
    for i, method in enumerate(sorted(methods.keys())):
        plt.subplot(1, 4, i + 1)
        plt.plot(x_train, y_train, '.')
        plt.gca().set_prop_cycle(None)
        predictions = methods[method].predict(x_test)
        if predictions.ndim < 2:
            plt.plot(x_test, predictions, '-')
        else:
            for quantile in predictions:
                plt.plot(x_test, quantile, '-')
        plt.gca().set_prop_cycle(None)
        for prob, quantile in zip(probs, z_test):
            plt.plot(x_test, quantile, '--', label="theoretical {0:0.2f}".format(prob))
        plt.title(method)
        plt.legend(fontsize=8)
    plt.show()

if __name__ == '__main__':
    main()



"""
:mod:`operalib.risk` implements risk model and their gradients.
"""
# Authors: Romain Brault <romain.brault@telecom-paristech.fr> with help from
#         the scikit-learn community.
#         Maxime Sangnier <maxime.sangnier@gmail.com>
# License: MIT

from numpy import inner, maximum, sum, multiply, subtract, where, dot, isclose
from numpy.linalg import norm
from numpy.ma import masked_invalid

from .preprocessing import SimplexCoding


class ORFFSCSVMLoss(object):
    """Define the Simplex Cone Support Vector Machine loss and its gradient."""

    def __init__(self):
        pass

    def __call__(self, coefs, ground_truth, phix, ker):
        pred = phix * coefs
        simplex_operator = SimplexCoding.code(ker.p + 1)
        pred = dot(pred.reshape((-1, ker.p)), simplex_operator)
        mask = isclose(dot(ground_truth.reshape((-1, ker.p)),
                           simplex_operator), 1)
        margin = maximum(0, 1. / ker.p + pred)
        margin[mask] = 0
        return margin.reshape((-1, ker.p + 1)).sum(axis=1).mean()

    def functional_grad(self, coefs, ground_truth, phix, ker):
        pred = phix * coefs
        simplex_operator = SimplexCoding.code(ker.p + 1)
        pred = dot(pred.reshape((-1, ker.p)), simplex_operator)
        mask = isclose(dot(ground_truth.reshape((-1, ker.p)),
                           simplex_operator), 1)
        margin = (1. / ker.p + pred) > 0
        margin[mask] = 0
        margin = dot(margin.reshape((-1, ker.p + 1)), simplex_operator.T)
        return phix.rmatvec(margin.flat) / ground_truth.size * ker.p

    def functional_grad_val(self, coefs, ground_truth, phix, ker):
        pred = phix * coefs
        simplex_operator = SimplexCoding.code(ker.p + 1)
        pred = dot(pred.reshape((-1, ker.p)), simplex_operator)
        mask = isclose(dot(ground_truth.reshape((-1, ker.p)),
                           simplex_operator), 1)
        margin = maximum(0, 1. / ker.p + pred)
        margin[mask] = 0
        margin = margin.reshape((-1, ker.p + 1))
        return (margin.sum(axis=1).mean(),
                phix.rmatvec(dot(margin > 0, simplex_operator.T).flat) /
                ground_truth.size * ker.p)


class ORFFHingeLoss(object):
    """Define Hinge loss for ORFF models and its gradient."""

    def __init__(self):
        """Initialize Hhnge ORFF loss.

        Parameters
        ==========
        """
        pass

    def __call__(self, coefs, ground_truth, phix, ker):
        """Compute the hinge loss value for ORFF models.


        Parameters
        ----------
        coefs : {vector-like}, shape = [n_samples1 * n_targets]
            Coefficient to optimise

        ground_truth : {vector-like}
            Targets samples

        phix : {LinearOperator}
            X ORFF mapping operator acting on the coefs

        Returns
        -------
        float : Empirical ORFF hinge loss

        """
        pred = phix * coefs
        return maximum(pred.reshape((-1, ker.p)).T -
                       sum(multiply(ground_truth, pred).reshape((-1, ker.p)),
                           axis=1).flat + 1 - ground_truth.reshape((-1,
                                                                    ker.p)).T,
                       0).sum(axis=0).mean()

    def functional_grad(self, coefs, ground_truth, phix, ker):
        pred = phix * coefs
        mask = (pred.reshape((-1, ker.p)).T -
                sum(multiply(ground_truth, pred).reshape((-1, ker.p)),
                    axis=1).flat + 1 - ground_truth.reshape((-1,
                                                             ker.p)).T > 0).T
        temp = multiply(mask.sum(axis=1).flat,
                        ground_truth.reshape((-1, ker.p)).T).T.flat
        mask = multiply(mask, 1 - ground_truth.reshape((-1, ker.p))).flat
        temp = phix.rmatvec(subtract(mask, temp))
        return (temp / ground_truth.size * ker.p)

    def functional_grad_val(self, coefs, ground_truth, phix, ker):
        pred = phix * coefs
        mrg = (pred.reshape((-1, ker.p)).T -
               sum(multiply(ground_truth, pred).reshape((-1, ker.p)),
                   axis=1).flat + 1 - ground_truth.reshape((-1, ker.p)).T).T
        val = maximum(mrg, 0).sum(axis=1).mean()
        mrg = mrg > 0
        temp = multiply(mrg.sum(axis=1).flat,
                        ground_truth.reshape((-1, ker.p)).T).T.flat
        mrg = multiply(mrg, 1 - ground_truth.reshape((-1, ker.p))).flat
        temp = phix.rmatvec(subtract(mrg, temp))
        return (val, temp / ground_truth.size * ker.p)


class ORFFLSLoss(object):
    """Define Least squares loss for ORFF models and its gradient."""

    def __init__(self):
        """Initialize Least squares ORFF loss.

        Parameters
        ==========
        """
        pass

    def __call__(self, coefs, ground_truth, phix, ker):
        """Compute the Least squares loss value for ORFF models.


        Parameters
        ----------
        coefs : {vector-like}, shape = [n_samples1 * n_targets]
            Coefficient to optimise

        ground_truth : {vector-like}
            Targets samples

        phix : {LinearOperator}
            X ORFF mapping operator acting on the coefs

        Returns
        -------
        float : Empirical ORFF least square loss

        """
        pred = phix * coefs
        res = pred - ground_truth
        return norm(res) ** 2 / (2 * ground_truth.size)

    def functional_grad(self, coefs, ground_truth, phix, ker):
        """Compute the Least squares loss gradient for ORFF models.


        Parameters
        ----------
        coefs : {vector-like}, shape = [n_samples1 * n_targets]
            Coefficient to optimise

        ground_truth : {vector-like}
            Targets samples

        phix : {LinearOperator}
            X ORFF mapping operator acting on the coefs

        Returns
        -------
        float : Empirical ORFF least square loss gradient

        """
        pred = phix * coefs
        res = pred - ground_truth
        return phix.rmatvec(res) / ground_truth.size

    def functional_grad_val(self, coefs, ground_truth, phix, ker):
        """Compute the Least squares loss gradient and value for ORFF models.


        Parameters
        ----------
        coefs : {vector-like}, shape = [n_samples1 * n_targets]
            Coefficient to optimise

        ground_truth : {vector-like}
            Targets samples

        phix : {LinearOperator}
            X ORFF mapping operator acting on the coefs

        Returns
        -------
        tuple : Empirical ORFF least square loss value and gradient

        """
        pred = phix * coefs
        res = pred - ground_truth
        return norm(res) ** 2 / (2 * ground_truth.size), \
            phix.rmatvec(res) / ground_truth.size


class ORFFRidgeRisk(object):
    """Define Ridge risk for ORFF models and its gradient."""

    def __init__(self, lbda, loss='LS'):
        """Initialize Empirical ORFF ridge risk.

        Parameters
        ----------
        lbda : {float}
            Small positive values of lbda improve the conditioning of the
            problem and reduce the variance of the estimates. Lbda corresponds
            to ``(2*C)^-1`` in other linear models such as LogisticRegression
            or LinearSVC.
        """
        self.lbda = lbda
        if loss is 'LS':
            self.loss = ORFFLSLoss()
        elif loss is 'Hinge':
            self.loss = ORFFHingeLoss()
        elif loss is 'SCSVM':
            self.loss = ORFFSCSVMLoss()
        else:
            raise NotImplementedError('unsupported loss')

    def __call__(self, coefs, ground_truth, phix, ker):
        """Compute the Empirical ORFF ridge risk.

        Parameters
        ----------
        coefs : {vector-like}, shape = [n_samples1 * n_targets]
            Coefficient to optimise

        ground_truth : {vector-like}
            Targets samples

        phix : {LinearOperator}
            X ORFF mapping operator acting on the coefs

        Returns
        -------
        float : Empirical ORFF ridge risk
        """
        return (self.loss(coefs, ground_truth, phix, ker) +
                self.lbda * norm(coefs) ** 2 / (2 * ground_truth.size))

    def functional_grad(self, coefs, ground_truth, phix, ker):
        """Compute the gradient of the Empirical ORFF ridge risk.

        Parameters
        ----------
        coefs : {vector-like}, shape = [n_samples1 * n_targets]
            Coefficient to optimise

        ground_truth : {vector-like}
            Targets samples

        phix : {LinearOperator}
            X ORFF mapping operator acting on the coefs

        Returns
        -------
        {vector-like} : gradient of the Empirical ORFF ridge risk
        """
        return (self.loss.functional_grad(coefs, ground_truth, phix, ker) +
                self.lbda * coefs / ground_truth.size)

    def functional_grad_val(self, coefs, ground_truth, phix, ker):
        """Compute the gradient and value of the Empirical ORFF ridge risk.

        Parameters
        ----------
        coefs : {vector-like}, shape = [n_samples1 * n_targets]
            Coefficient to optimise

        ground_truth : {vector-like}
            Targets samples

        phix : {LinearOperator}
            X ORFF mapping operator acting on the coefs

        Returns
        -------
        Tuple{float, vector-like} : Empirical ORFF ridge risk and its gradient
        returned as a tuple.
        """
        val_loss, grad_loss = self.loss.functional_grad_val(coefs,
                                                            ground_truth,
                                                            phix, ker)
        return (val_loss +
                self.lbda * norm(coefs) ** 2 / (2 * ground_truth.size),
                grad_loss + self.lbda * coefs / ground_truth.size)


class OVKRidgeRisk(object):
    """Define Kernel ridge risk and its gradient."""

    def __init__(self, lbda):
        """Initialize Empirical kernel ridge risk.

        Parameters
        ----------
        lbda : {float}
            Small positive values of lbda improve the conditioning of the
            problem and reduce the variance of the estimates.  Lbda corresponds
            to ``(2*C)^-1`` in other linear models such as LogisticRegression
            or LinearSVC.
        """
        self.lbda = lbda

    def __call__(self, coefs, ground_truth, Gram,
                 weight=None, zeronan=None):
        """Compute the Empirical OVK ridge risk.

        Parameters
        ----------
        coefs : {vector-like}, shape = [n_samples1 * n_targets]
            Coefficient to optimise

        ground_truth : {vector-like}
            Targets samples

        Gram : {LinearOperator}
            Gram matrix acting on the coefs

        weight: {LinearOperator}

        zeronan: {LinearOperator}

        Returns
        -------
        float : Empirical OVK ridge risk
        """
        np = ground_truth.size
        pred = Gram * coefs
        reg = inner(coefs, pred)  # reg in rkhs
        vgt = masked_invalid(ground_truth)
        vgt[where(vgt.mask)] = pred[where(vgt.mask)]
        if weight is None or zeronan is None:
            obj = norm(pred - vgt) ** 2 / (2 * np)
        else:
            wpred = weight * pred  # sup x identity | unsup x lbda_m x L
            res = zeronan * (wpred - vgt)
            wip = wpred - zeronan * wpred  # only unsup part of wpred
            lap = inner(wip, pred)  # Laplacian part x lambda_m

            obj = norm(zeronan * res) ** 2 / (2 * np)  # Loss
            obj += lap / (2 * np)  # Laplacian regularization
        obj += self.lbda * reg / (2 * np)  # Regulariation
        return obj

    def functional_grad(self, coefs, ground_truth, Gram,
                        weight=None, zeronan=None):
        """Compute the gradient of the Empirical OVK ridge risk.

        Parameters
        ----------
        coefs : {vector-like}, shape = [n_samples1 * n_targets]
            Coefficient to optimise

        ground_truth : {vector-like}
            Targets samples

        Gram : {LinearOperator}
            Gram matrix acting on the coefs

        weight: {LinearOperator}

        zeronan: {LinearOperator}

        Returns
        -------
        {vector-like} : gradient of the Empirical OVK ridge risk
        """
        np = ground_truth.size
        pred = Gram * coefs
        vgt = masked_invalid(ground_truth)
        vgt[where(vgt.mask)] = pred[where(vgt.mask)]
        if weight is None or zeronan is None:
            res = pred - vgt
        else:
            res = weight * pred - zeronan * vgt
        return Gram * res / np + self.lbda * pred / np

    def functional_grad_val(self, coefs, ground_truth, Gram,
                            weight=None, zeronan=None):
        """Compute the gradient and value of the Empirical OVK ridge risk.

        Parameters
        ----------
        coefs : {vector-like}, shape = [n_samples1 * n_targets]
            Coefficient to optimise

        ground_truth : {vector-like}
            Targets samples

        Gram : {LinearOperator}
            Gram matrix acting on the coefs

        L : array, shape = [n_samples_miss, n_samples_miss]
            Graph Laplacian of data with missing targets (semi-supervised
            learning).

        Returns
        -------
        Tuple{float, vector-like} : Empirical OVK ridge risk and its gradient
        returned as a tuple.
        """
        np = ground_truth.size
        pred = Gram * coefs
        vgt = masked_invalid(ground_truth)
        vgt[where(vgt.mask)] = pred[where(vgt.mask)]
        reg = inner(coefs, pred)  # reg in rkhs
        if weight is None or zeronan is None:
            res = pred - vgt
            obj = norm(res) ** 2 / (2 * np)
        else:
            wpred = weight * pred  # sup x identity | unsup x lbda_m x L
            res = wpred - zeronan * vgt
            wip = wpred - zeronan * wpred  # only unsup part of wpred
            lap = inner(wip, pred)  # Laplacian part x lambda_m

            obj = norm(zeronan * res) ** 2 / (2 * np)  # Loss
            obj += lap / (2 * np)  # Laplacian regularization
        obj += self.lbda * reg / (2 * np)  # Regulariation
        return obj, Gram * res / np + self.lbda * pred / np



"""
:mod:`operalib.ridge` implements Operator-Valued Kernel ridge
regression.
"""
# Author: Romain Brault <romain.brault@telecom-paristech.fr> with help from
#         the scikit-learn community.
#         Maxime Sangnier <maxime.sangnier@gmail.com>
# License: MIT

from scipy.optimize import minimize
from scipy.sparse.linalg import LinearOperator
from numpy import (reshape, eye, zeros, empty, dot, all, isnan, diag, number,
                   issubdtype)

from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils import check_array
from sklearn.utils.validation import check_is_fitted
from sklearn.metrics.pairwise import rbf_kernel

from .metrics import first_periodic_kernel
from .kernels import DecomposableKernel, RBFCurlFreeKernel
from .risk import OVKRidgeRisk
from .signal import get_period

# When adding a new kernel, update this table and the _get_kernel_map method
PAIRWISE_KERNEL_FUNCTIONS = {
    'DGauss': DecomposableKernel,
    'DPeriodic': DecomposableKernel,
    'CurlF': RBFCurlFreeKernel}


def _graph_Laplacian(similarities):
    return diag(similarities.sum(axis=1)) - similarities


class _SemisupLinop:

    def __init__(self, lbda2, is_sup, L, p):
        self.lbda2 = lbda2
        self.is_sup = is_sup.ravel()
        self.L = L
        self.p = p
        self.ns = is_sup.shape[0] - L.shape[0]
        self.ls = L.shape[0]

    def _dot_U(self, vec):
        mat = vec.reshape((self.ns + self.ls, self.p))
        res = empty((self.ns + self.ls, self.p))
        res[self.is_sup, :] = mat[self.is_sup, :]
        res[~self.is_sup, :] = self.lbda2 * dot(self.L, mat[~self.is_sup, :])

        return res.ravel()

    def _dot_JT(self, vec):
        mat = vec.reshape((self.ns + self.ls, self.p))
        res = empty((self.ns + self.ls, self.p))
        res[self.is_sup, :] = mat[self.is_sup, :]
        res[~self.is_sup, :] = 0

        return res.ravel()

    def gen(self):
        shape_U = ((self.ns + self.ls) * self.p, (self.ns + self.ls) * self.p)
        shape_JT = ((self.ns + self.ls) * self.p, (self.ns + self.ls) * self.p)

        # return U, JT
        return (LinearOperator(shape_U,
                               dtype=self.L.dtype,
                               matvec=lambda b: self._dot_U(b),
                               rmatvec=lambda b: self._dot_U(b)),
                LinearOperator(shape_JT,
                               dtype=self.L.dtype,
                               matvec=lambda b: self._dot_JT(b),
                               rmatvec=lambda b: self._dot_JT(b)))


class OVKRidge(BaseEstimator, RegressorMixin):
    """Operator-Valued kernel ridge regression.

    Operator-Valued kernel ridge regression (OVKRR) combines ridge regression
    (linear least squares with l2-norm regularization) with the (OV)kernel
    trick. It learns a linear function in the space induced by the
    respective kernel and the data. For non-linear kernels, this corresponds to
    a non-linear function in the original space.

    The form of the model learned by OVKRR is identical to support vector
    regression (SVR). However, different loss functions are used: OVKRR uses
    squared error loss while support vector regression uses epsilon-insensitive
    loss, both combined with l2 regularization. In contrast to SVR, fitting a
    OVKRR model can be done in closed-form and is typically faster for
    medium-sized datasets. On the other  hand, the learned model is non-sparse
    and thus slower than SVR, which learns a sparse model for epsilon > 0, at
    prediction-time.

    Optimization problem solved for learning:
    .. math::
        \min_{h \in \mathcal H}~ \frac{\lambda}{2} \|h\|_{\mathcal H}^2 +
        \frac{1}{np} \sum_{i=1}^n \|y_j - h(x_i)\|_{\mathcal Y}^2 +
        \frac{\lambda_m}{2} \sum_{i=1}^n W_{ij}
        \|h(x_i) - h(x_j)\|_{\mathcal Y}^2

    Attributes
    ----------
    dual_coef_ : array, shape = [n_samples x n_targest]
        Weight vector(s) in kernel space

    linop_ : callable
        Callable which associate to the training points X the Gram matrix (the
        Gram matrix being a LinearOperator)

    A_ : array, shape = [n_targets, n_targets]
        Set when Linear operator used by the decomposable kernel is default or
        None.

    L_ : array, shape = [n_samples_miss, n_samples_miss]
        Graph Laplacian of data with missing targets (semi-supervised
        learning).

    period_ : float
        Set when period used by the First periodic kernel is 'autocorr'.

    solver_res_ : any
        Raw results returned by the solver.

    References
    ----------
    * Micchelli, Charles A., and Massimiliano Pontil.
      "On learning vector-valued functions." Neural computation
      17.1 (2005): 177-204.

    * Alvarez, Mauricio A., Lorenzo Rosasco, and Neil D. Lawrence.
      "Kernels for vector-valued functions: A review." arXiv preprint
      arXiv:1106.6251 (2011). APA

    * Brouard Celine, d'Alche-Buc Florence and Szafranski Marie.
      "Input Output Kernel Regression," Hal preprint
      hal-01216708 (2015).


    See also
    --------
    sklearn.OVKRidge
        Linear ridge regression.
    sklearn.KernelOVKRidge
        Kernel ridge regression.
    sklearn.SVR
        Support Vector Regression implemented using libsvm.

    Examples
    --------
    >>> import operalib as ovk
    >>> import numpy as np
    >>> n_samples, n_features, n_targets = 10, 5, 5
    >>> rng = np.random.RandomState(0)
    >>> y = rng.randn(n_samples, n_targets)
    >>> X = rng.randn(n_samples, n_features)
    >>> clf = ovk.OVKRidge('DGauss', lbda=1.0)
    >>> clf.fit(X, y)
    # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
    OVKRidge(...)
    """

    def __init__(self,
                 ovkernel='DGauss', lbda=1e-5, lbda_m=0.,
                 A=None, gamma=None, gamma_m=None, theta=0.7,
                 period='autocorr', autocorr_params=None,
                 solver='L-BFGS-B', solver_params=None):
        """Initialize OVK ridge regression model.

        Parameters
        ----------

        ovkernel : {string, callable}, default='DGauss'
            Kernel mapping used internally. A callable should accept two
            arguments, and should return a LinearOperator.

        lbda : {float}, default=1e-5
            Small positive values of lbda improve the conditioning of the
            problem and reduce the variance of the estimates.  Lbda corresponds
            to ``(2*C)^-1`` in other linear models such as LogisticRegression
            or LinearSVC.

        lbda_m : {float}, default=0.
            Regularization parameter for quadratic penalty on data with missing
            targets.

        A : {LinearOperator, array-like, sparse matrix}, default=None
            Linear operator used by the decomposable kernel. If default is
            None, wich is set to identity matrix of size y.shape[1] when
            fitting.

        gamma : {float}, default=None.
            Gamma parameter for the Decomposable Gaussian kernel.
            Ignored by other kernels.

        gamma_m : {float}, default=None.
            Gamma parameter for the graph Laplacian inducing penalty on data
            with missing targets.

        theta : {float}, default=.7
            Theta parameter for the Decomposable First Periodic kernel.
            Ignored by other kernels.

        period : {float}, default=default_period
            Period parameter for the First periodic kernel. If optional modules
            have been imported then default_period is 2 * pi. Otherwise it uses
            autocorrelation methods to determine the period.

        solver : {callable}, default=scipy.optimize.fmin_l_bfgs_b
            Solver able to find the minimum of the ridge problem.
            scipy.optimize.fmin_l_bfgs_b(*solver_params)[0] must return the
            optimal solution.

        autocorr_params : {mapping of string to any}
            Additional parameters (keyword arguments) for the period detection
            for periodic kernels. If None, parameter choice is left to the
            period detection method.

        solver_params : {mapping of string to any}, optional
            Additional parameters (keyword arguments) for solver function
            passed as callable object.
        """
        self.ovkernel = ovkernel
        self.lbda = lbda
        self.lbda_m = lbda_m
        self.A = A
        self.gamma = gamma
        self.gamma_m = gamma_m
        self.theta = theta
        self.period = period
        self.autocorr_params = autocorr_params
        self.solver = solver
        self.solver_params = solver_params

    def _validate_params(self):
        # check on self.ovkernel is performed in method __get_kernel
        if self.lbda < 0:
            raise ValueError('lbda must be positive')
        if self.lbda_m < 0:
            raise ValueError('lbda_m must be positive')
        # if self.A < 0: # Check whether A is S PD would be really expensive
        #     raise ValueError('A must be a symmetric positive operator')
        if self.gamma is not None:
            if self.gamma < 0:
                raise ValueError('sigma must be positive or default (None)')
        if self.theta < 0:
            raise ValueError('theta must be positive')
        if isinstance(self.period, (int, float)):
            if self.period < 0:
                raise ValueError('period must be positive')
        # TODO, add supported solver check

    def _default_decomposable_op(self, y):
        # TODO: check NaN values (semi-sup learning)
        if self.A is not None:
            return self.A
        elif y.ndim == 2:
            return eye(y.shape[1])
        else:
            return eye(1)

    def _default_period(self, X, y):
        if self.period is 'autocorr':
            autocorr_params = self.autocorr_params or {}
            return get_period(X, y, **autocorr_params)
        elif isinstance(self.period, (int, float)):
            return self.period
        else:
            raise ValueError('period must be a positive number or a valid '
                             'string')

    def _get_kernel_map(self, X, y):
        # When adding a new kernel, update this table and the _get_kernel_map
        # method
        if callable(self.ovkernel):
            ovkernel = self.ovkernel
        elif type(self.ovkernel) is str:
            # 1) check string and assign the right parameters
            if self.ovkernel == 'DGauss':
                self.A_ = self._default_decomposable_op(y)
                kernel_params = {'A': self.A_, 'scalar_kernel': rbf_kernel,
                                 'scalar_kernel_params': {'gamma': self.gamma}}
            elif self.ovkernel == 'DPeriodic':
                self.A_ = self._default_decomposable_op(y)
                self.period_ = self._default_period(X, y)
                kernel_params = {'A': self.A_,
                                 'scalar_kernel': first_periodic_kernel,
                                 'scalar_kernel_params': {'gamma': self.theta,
                                                          'period':
                                                          self.period_}, }
            elif self.ovkernel == 'CurlF':
                kernel_params = {'gamma': self.gamma}
            else:
                raise NotImplementedError('unsupported kernel')
            # 2) Uses lookup table to select the right kernel from string
            ovkernel = PAIRWISE_KERNEL_FUNCTIONS[self.ovkernel](
                **kernel_params)
        else:
            raise NotImplementedError('unsupported kernel')
        return ovkernel(X)

    def _decision_function(self, X):
        pred = self.linop_(X) * self.dual_coefs_

        return reshape(pred, (X.shape[0], self.linop_.p)) \
            if self.linop_.p > 1 else pred

    def fit(self, X, y):
        """Fit OVK ridge regression model.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training data.

        y : {array-like}, shape = [n_samples] or [n_samples, n_targets]
            Target values. numpy.NaN for missing targets (semi-supervised
            learning).

        Returns
        -------
        self : returns an instance of self.
        """
        X = check_array(X, force_all_finite=True, accept_sparse=False,
                        ensure_2d=True)
        y = check_array(y, force_all_finite=False, accept_sparse=False,
                        ensure_2d=False)
        if y.ndim == 1:
            y = check_array(y, force_all_finite=True, accept_sparse=False,
                            ensure_2d=False)
        self._validate_params()

        solver_params = self.solver_params or {}

        self.linop_ = self._get_kernel_map(X, y)
        Gram = self.linop_(X)
        risk = OVKRidgeRisk(self.lbda)

        if not issubdtype(y.dtype, number):
            raise ValueError("Unknown label type: %r" % y.dtype)
        if y.ndim > 1:
            is_sup = ~all(isnan(y), axis=1)
        else:
            is_sup = ~isnan(y)

        if sum(~is_sup) > 0:
            self.L_ = _graph_Laplacian(rbf_kernel(X[~is_sup, :],
                                                  gamma=self.gamma_m))
        else:
            self.L_ = empty((0, 0))

        p = y.shape[1] if y.ndim > 1 else 1
        weight, zeronan = _SemisupLinop(self.lbda_m, is_sup, self.L_, p).gen()

        self.solver_res_ = minimize(risk.functional_grad_val,
                                    zeros(Gram.shape[1]),
                                    args=(y.ravel(), Gram, weight, zeronan),
                                    method=self.solver,
                                    jac=True,
                                    options=solver_params)
        self.dual_coefs_ = self.solver_res_.x
        return self

    def predict(self, X):
        """Predict using the OVK ridge model.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Samples.

        Returns
        -------
        C : {array}, shape = [n_samples] or [n_samples, n_targets]
            Returns predicted values.
        """
        check_is_fitted(self, ['dual_coefs_', 'linop_'], all_or_any=all)
        X = check_array(X, force_all_finite=True, accept_sparse=False,
                        ensure_2d=True)
        return self._decision_function(X)



"""OVK learning, unit tests.

The :mod:`sklearn.tests.awful` tests awful dataset.
"""
from numpy import isnan

import operalib as ovk


def test_awful():
    """Test awful function."""
    _, targets = ovk.toy_data_curl_free_field(n_samples=500)
    targets = ovk.datasets.awful(targets)
    assert isnan(targets).any()



"""
:mod:`operalib.quantile` implements joint quantile regression with
Operator-Valued Kernels.
"""
# Authors: Maxime Sangnier <maxime.sangnier@gmail.com>
#          Romain Brault <romain.brault@telecom-paristech.fr> with help from
#          the scikit-learn community.
# License: MIT
from numpy import (asarray, eye, reshape, argsort, kron, ones, diag, delete,
                   zeros, r_, c_, percentile, fmax)
from numpy import inf as npinf
from numpy import sum as npsum

from cvxopt import matrix, solvers

from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils import check_X_y, check_array
from sklearn.utils.validation import check_is_fitted, check_consistent_length
from sklearn.metrics.pairwise import rbf_kernel

# from operalib. .kernels import DecomposableKernel
from operalib.kernels import DecomposableKernel

# When adding a new kernel, update this table and the _get_kernel_map method
PAIRWISE_KERNEL_FUNCTIONS = {
    'DGauss': DecomposableKernel,
    'DPeriodic': DecomposableKernel, }

# pylint: disable=C0103,W0201,R0902,R0913


class Quantile(BaseEstimator, RegressorMixin):
    """Joint quantile regression with operator-valued kernels.
    Joint quantile regression enables to learn and predict simultaneously
    several conditional quantiles (for prescribed quantile levels). Following
    the seminal work by Koenker and Basset (1978), the method minimizes an
    empirical risk based on the pinball loss and a squared-norm regularization.
    The hypothesis space considered here is the reproducing kernel Hilbert
    space generated by an operator-valued kernel chosen by the practitioner.

    Attributes
    ----------
    model : dict
        coefs : {array-like}, shape [n_samples*n_probs, 1]
            Dual coefficients of the kernel machine
        intercept : {array-like}, shape [n_probs]
            Vector of intercepts

    References
    ----------
    * R. Koenker and G. Bassett
      "Regression Quantiles", Econometrica, 46(1):33-50, 1978.
    * R. Koenker
      "Quantile Regression", Cambridge University Press, Cambridge, New York,
      2005.
    * M. Sangnier, O. Fercoq and F. d'Alche-Buc
      "Joint quantile regression in vector-valued RKHSs." In Advances in Neural
    See also
    --------

    operalib.Ridge
        Operator-Valued kernel ridge regression
    sklearn.Ridge
        Linear ridge regression.
    sklearn.KernelRidge
        Kernel ridge regression.
    sklearn.SVR
        Support Vector Regression implemented using libsvm.

    Examples
    --------
    >>> import operalib as ovk
    >>> import numpy as np
    >>> n_samples, n_features = 10, 5
    >>> quantile_levels = [0.1, 0.5, 0.9]
    >>> rng = np.random.RandomState(0)
    >>> y = rng.randn(n_samples)
    >>> X = rng.randn(n_samples, n_features)
    >>> reg = ovk.Quantile('DGauss', lbda=1.0)
    >>> reg.fit(X, y)  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
    Quantile(gamma=None, gamma_quantile=0.0, kernel='DGauss',
        kernel_params=None, lbda=1.0, nc_const=False, probs=0.5, tol=None,
        verbose=False)
    """

    def __init__(self, kernel='DGauss', probs=0.5, lbda=1e-5, gamma=None,
                 gamma_quantile=0., eps=0., tol=None, nc_const=False,
                 kernel_params=None, verbose=False):
        """Initialize quantile regression model.

        Parameters
        ----------
        probs : {float, list}, default=0.5
            Probabilities (quantiles levels).
        kernel : {string, callable}, default='DGauss'
            Kernel mapping used internally. A callable should accept two
            arguments and the keyword arguments passed to this object as
            kernel_params, and should return a LinearOperator.
        lbda : {float}, default=1e-5
            Regularization parameter.  Lbda corresponds to ``C^-1`` in other
            large margin models such as SVC.
        gamma : {float}, default=None.
            Gamma parameter for the Decomposable Gaussian kernel.
            Ignored by other kernels.
        gamma_quantile : {float}, default=None.
            Gamma parameter for the output Gaussian kernel.
        eps : {float}, default=0
            Threshold for the epsilon-loss (data sparse regression, only
            without non-crossing constraints)
        tol : {float}, default=None
            Optimization tolerance (None leads to the default value of the
            solver).
        nc_const : {Boolean}, default=False
            Whether to enforce non-crossing constraints when optimizing
            (default is no).
        kernel_params : {mapping of string to any}, optional
            Additional parameters (keyword arguments) for kernel function
            passed as callable object.
        verbose : {Boolean}, default=False
            Verbosity
        """
        self.probs = probs
        self.lbda = lbda
        self.kernel = kernel
        self.gamma = gamma
        self.gamma_quantile = gamma_quantile
        self.eps = eps
        self.tol = tol
        self.nc_const = nc_const
        self.kernel_params = kernel_params
        self.verbose = verbose

        # self.linop_ = None
        # self.reg_c_ = None
        # self.sol_ = None
        # self.model_ = None

    def _validate_params(self):
        # check on self.kernel is performed in method __get_kernel
        if self.lbda < 0:
            raise ValueError('lbda must be positive')
        # if self.A < 0: # Check whether A is S PD would be really expensive
        #     raise ValueError('A must be a symmetric positive operator')
        if self.gamma is not None:
            if self.gamma < 0:
                raise ValueError('sigma must be positive or default (None)')
        # probs = asarray(self.probs).reshape(-1, 1)
        probs = asarray(self.probs).flatten()
        if (probs < 0).any() or (probs > 1).any():
            raise ValueError('Probabilities must be in [0., 1.]')

    def _default_decomposable_op(self):
        probs = asarray(self.probs).reshape((1, -1))  # 2D array
        return (rbf_kernel(probs.T, gamma=self.gamma_quantile)
                if self.gamma_quantile != npinf else eye(len(self.probs)))

    def _get_kernel_map(self, inputs):
        # When adding a new kernel, update this table and the _get_kernel_map
        # method
        if callable(self.kernel):
            kernel_params = self.kernel_params or {}
            ov_kernel = self.kernel(**kernel_params)
        elif isinstance(self.kernel, str):
            # 1) check string and assign the right parameters
            if self.kernel == 'DGauss':
                kernel_params = {'A': self._default_decomposable_op(),
                                 'scalar_kernel': rbf_kernel,
                                 'scalar_kernel_params': {'gamma': self.gamma}}
            else:
                raise NotImplementedError('unsupported kernel')
            # 2) Uses lookup table to select the right kernel from string
            ov_kernel = PAIRWISE_KERNEL_FUNCTIONS[self.kernel](**kernel_params)
        else:
            raise NotImplementedError('unsupported kernel')
        return ov_kernel(inputs)

    def _decision_function(self, X):
        n_samples = X.shape[0]
        n_quantiles = asarray(self.probs).reshape((-1, 1)).size

        pred = reshape(self.linop_(X) * self.model_['coefs'],
                       (n_samples, n_quantiles))
        pred += asarray(self.model_['intercept']).squeeze()

        return pred.T if self.linop_.p > 1 else pred.T.ravel()

    def predict(self, X):
        """Predict conditional quantiles.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Samples.

        Returns
        -------
        y : {array}, shape = [n_samples, n_quantiles]
            Returns predicted values for each prescribed quantile level.
        """
        check_is_fitted(self, ['model_', 'linop_'], all_or_any=all)
        X = check_array(X)
        y = self._decision_function(X)
        y[y < 1e-9] = 0 # Stability hack
        return y

    def fit(self, X, y):
        """Fit joint quantile regression model.

        Parameters
        ----------
        inputs : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training data.
        targets : {array-like}, shape = [n_samples]
            Target values.

        Returns
        -------
        self : returns an instance of self.
        """
        if self.eps > 0 and self.nc_const:
            raise UserWarning("eps is considered null because you chose to "
                              "enfoce non-crossing constraints.")
        X, y = check_X_y(X, y, ['csr', 'csc', 'coo'], y_numeric=True)
        y = asarray(y).flatten()
        self._validate_params()

        self.linop_ = self._get_kernel_map(X)
        gram = self.linop_.Gram_dense(X)
        self.reg_c_ = 1. / self.lbda

        # Solve the optimization problem
        # probs = asarray(self.probs).reshape((-1, 1))
        probs = asarray(self.probs).flatten()
        if self.nc_const:
            self._qp_nc(gram, y, probs)
        else:
            self._coneqp(gram, y, probs)
        return self

    def _qp_nc(self, gram, targets, probs):
        # Needed to sort constraints on quantiles levels
        ind = argsort(probs.T).squeeze()

        n_quantiles = probs.size  # Number of quantiles to predict
        n_coefs = gram.shape[0]  # Number of coefficients
        n_train = int(n_coefs / n_quantiles)  # Number of training instances
        # Number of non-crossing dual variables
        n_no_cross = n_train * (n_quantiles - 1)
        probs = kron(ones(n_train), probs.squeeze())  # Quantiles levels

        # Difference matrix
        diff_mat = -eye(n_quantiles) + diag(ones(n_quantiles - 1), 1)
        diff_mat = delete(diff_mat, -1, 0)
        diff_mat = diff_mat.T[ind].T
        # Quad. part of the objective function
        gram = matrix(r_[c_[gram, zeros((n_coefs, n_no_cross))],
                         zeros((n_no_cross, n_coefs + n_no_cross))])
        # Linear part of the objective
        q_lin = matrix(r_[-kron(targets, ones(n_quantiles)),
                          zeros(n_no_cross)])
        # LHS of the inequality constraint
        g_lhs = matrix(r_[c_[eye(n_coefs), -kron(eye(n_train), diff_mat.T)],
                          c_[-eye(n_coefs), kron(eye(n_train), diff_mat.T)],
                          c_[zeros((n_no_cross, n_coefs)),
                             -eye(n_no_cross)]])
        # RHS of the inequality constraint
        h_rhs = matrix(r_[self.reg_c_ * probs,
                          self.reg_c_ * (1 - probs),
                          zeros(n_train * (n_quantiles - 1))])
        # LHS of the equality constraint
        lhs_eqc = matrix(c_[kron(ones(n_train), eye(n_quantiles)),
                            zeros((n_quantiles, n_no_cross))])

        solvers.options['show_progress'] = self.verbose
        if self.tol:
            solvers.options['reltol'] = self.tol
        # Solve the dual optimization problem
        self.sol_ = solvers.qp(gram, q_lin, g_lhs, h_rhs, lhs_eqc,
                               matrix(zeros(n_quantiles)))

        self.model_ = {'coefs': asarray(self.sol_['x'][:n_coefs]),
                       'intercept': asarray(self.sol_['y']).squeeze()}

    def _qp(self, gram, targets, probs):
        n_quantiles = probs.size  # Number of quantiles to predict
        n_coefs = gram.shape[0]  # Number of variables
        # Quantiles levels
        kronprobs = kron(ones(int(n_coefs / n_quantiles)), probs.squeeze())

        # Quadratic part of the objective
        gram = matrix(gram)
        # Linear part of the objective
        q_lin = matrix(-kron(targets, ones(n_quantiles)))

        # LHS of the inequality constraint
        g_lhs = matrix(r_[eye(n_coefs), -eye(n_coefs)])
        # RHS of the inequality
        h_rhs = matrix(r_[self.reg_c_ * kronprobs,
                          self.reg_c_ * (1 - kronprobs)])
        # LHS of the equality constraint
        lhs_eqc = matrix(kron(ones(int(n_coefs / n_quantiles)),
                              eye(n_quantiles)))

        solvers.options['show_progress'] = self.verbose
        if self.tol:
            solvers.options['reltol'] = self.tol
        # Solve the dual optimization problem
        self.sol_ = solvers.qp(gram, q_lin, g_lhs, h_rhs, lhs_eqc,
                               matrix(zeros(n_quantiles)))

        # Set coefs
        coefs = asarray(self.sol_['x'])

        # Set the intercept

        # Erase the previous intercept before prediction
        self.model_ = {'coefs': coefs, 'intercept': 0}
        predictions = self.predict(self.linop_.X)
        if predictions.ndim < 2:
            predictions = predictions.reshape(1, -1)  # 2D array
        intercept = [percentile(targets - pred, 100. * prob) for
                     (pred, prob) in zip(predictions, probs)]
        intercept = asarray(intercept).squeeze()
        self.model_ = {'coefs': coefs, 'intercept': intercept}

    def _coneqp(self, gram, targets, probs):
        n_quantiles = probs.size  # Number of quantiles to predict
        n_coefs = gram.shape[0]  # Number of variables
        n_samples = n_coefs // n_quantiles
        # Quantiles levels
        kronprobs = kron(ones(int(n_coefs / n_quantiles)), probs.squeeze())

        solvers.options['show_progress'] = self.verbose
        if self.tol:
            solvers.options['reltol'] = self.tol

        if self.eps == 0:
            # Quadratic part of the objective
            gram = matrix(gram)
            # Linear part of the objective
            q_lin = matrix(-kron(targets, ones(n_quantiles)))

            # LHS of the inequality constraint
            g_lhs = matrix(r_[eye(n_coefs), -eye(n_coefs)])
            # RHS of the inequality
            h_rhs = matrix(r_[self.reg_c_ * kronprobs,
                              self.reg_c_ * (1 - kronprobs)])
            # LHS of the equality constraint
            lhs_eqc = matrix(kron(ones(n_samples), eye(n_quantiles)))

            # Solve the dual optimization problem
            self.sol_ = solvers.qp(gram, q_lin, g_lhs, h_rhs, lhs_eqc,
                                   matrix(zeros(n_quantiles)))

            # Set coefs
            coefs = asarray(self.sol_['x'])
        else:
            def build_lhs(m, p):
                # m: n_samples
                # p: n_quantiles
                n = m*p  # n_variables

                # Get the norm bounds (m last variables)
                A = zeros(p+1)
                A[0] = -1
                A = kron(eye(m), A).T
                # Get the m p-long vectors
                B = kron(eye(m), c_[zeros(p), eye(p)].T)
                # Box constraint
                C = c_[r_[eye(n), -eye(n)], zeros((2*n, m))]
                # Set everything together
                C = r_[C, c_[B, A]]
                return C

            # Quadratic part of the objective
            gram = matrix(r_[c_[gram, zeros((n_coefs, n_samples))],
                             zeros((n_samples, n_coefs+n_samples))])
            # Linear part of the objective
            q_lin = matrix(r_[-kron(targets, ones(n_quantiles)),
                              ones(n_samples)*self.eps])

            # LHS of the inequality constraint
            g_lhs = matrix(build_lhs(n_samples, n_quantiles))
            # RHS of the inequality
            h_rhs = matrix(r_[self.reg_c_ * kronprobs,
                              self.reg_c_ * (1-kronprobs),
                              zeros(n_samples * (n_quantiles+1))])
            # LHS of the equality constraint
            lhs_eqc = matrix(c_[kron(ones(n_samples),
                                eye(n_quantiles)),
                                zeros((n_quantiles, n_samples))])
            # Parameters of the optimization problem
            dims = {'l': 2*n_coefs, 'q': [n_quantiles+1]*n_samples, 's': []}

            # Solve the dual optimization problem
            self.sol_ = solvers.coneqp(gram, q_lin, g_lhs, h_rhs, dims,
                                       lhs_eqc, matrix(zeros(n_quantiles)))

            # Set coefs
            coefs = asarray(self.sol_['x'][:n_coefs])

        # Set the intercept

        # Erase the previous intercept before prediction
        self.model_ = {'coefs': coefs, 'intercept': 0}
        predictions = self.predict(self.linop_.X)
        if predictions.ndim < 2:
            predictions = predictions.reshape(1, -1)  # 2D array
        intercept = [percentile(targets - pred, 100. * prob) for
                     (pred, prob) in zip(predictions, probs)]
        intercept = asarray(intercept).squeeze()
        self.model_ = {'coefs': coefs, 'intercept': intercept}

    @staticmethod
    def pinball_loss(y_true, y_pred, probs):
        """Compute the pinball loss.

        Parameters
        ----------
        pred : {array-like}, shape = [n_quantiles, n_samples] or [n_samples]
            Predictions.
        y : {array-like}, shape = [n_samples]
            Targets.

        Returns
        -------
        l : {array}, shape = [n_quantiles]
            Average loss for each quantile level.
        """
        probs = asarray(probs).reshape(-1)
        check_consistent_length(y_true, y_pred.T)
        y_true = check_array(y_true.reshape((-1, 1)),
                             ensure_2d=True)
        y_pred = check_array(y_pred.T.reshape((y_true.shape[0], -1)),
                             ensure_2d=True)
        residual = y_true - y_pred
        loss = npsum([fmax(prob * res, (prob - 1) * res) for (res, prob) in
                      zip(residual.T, probs)], axis=1)
        return loss / y_true.size

    def score(self, X, y, sample_weight=None):
        """Compute the pinball score for the given dataset.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training data.
        y : {array-like}, shape = [n_samples]
            Target values.

        Returns
        -------
        l : {float}
            Average pinball score (the higher, the better).
        """
        check_is_fitted(self, ['model_', 'linop_'], all_or_any=all)
        X, y = check_X_y(X, y)
        return 1 - Quantile.pinball_loss(y, self.predict(X), self.probs).mean()



"""Synthetic datasets for quantile learning."""

from scipy.stats import norm
from numpy import sort, sin, array, pi

from sklearn.utils import check_random_state
from sklearn.base import BaseEstimator


class SinePattern(BaseEstimator):
    """"Sine pattern.

    The inputs X is drawn uniformly in the 1D interval [min, max].  The target
    y is computed as a sine wave with some period modulated by a sine wave
    envelope with another period and shifted by a constant.

    Attributes
    ----------
    inputs_bound : :rtype: (float, float)
        The bounds (min, max) of the inputs.

    sine_period : float
        The period of the sine wave.

    enveloppe : :rtype: (float, float)
        The (shift, period) of the enveloppe of the enveloppe.

    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    References
    ----------
    * Maxime Sangnier, Olivier Fercoq, Florence D'Alche-Buc
    "Joint quantile regression in vector-valued RKHSs." In Advances in Neural
    Information Processing Systems (2016, pp. 3693-3701).
    """

    def __init__(self, inputs_bound=(0., 1.5), sine_period=1.,
                 enveloppe=(1. / 3., 1.), random_state=None):
        """Initialize the sine pattern.

        Parameters
        ----------
        inputs_bound : :rtype: (float, float), optional (default = (0., 1.5))
            The bounds (min, max) of the inputs

        sine_period : float, optional (default = 1.0)
            The period of the sine wave.

        enveloppe : :rtype: (float, float), optional (default = (1. / 3., 1.))
            The (shift, period) of the enveloppe of the sine wave

        random_state : int, RandomState instance or None,
                       optional (default=None)
            If int, random_state is the seed used by the random number
            generator; If RandomState instance, random_state is the random
            number generator; If None, the random number generator is the
            RandomState instance used by `np.random`.
        """
        self.inputs_bound = inputs_bound
        self.sine_period = sine_period
        self.enveloppe = enveloppe
        self.random_state = random_state

    def __call__(self, n_samples):
        """Generate the data.

        Parameters
        ----------

        n_points : int
            The number of points in the generated dataset.

        Returns
        -------

        inputs : array, shape = [n_samples, 1]
            Returns the generated inputs.

        targets : array, shape = [n_samples, 1]
            Returns the generated targets.
        """
        random_state = check_random_state(self.random_state)
        envelope_period, envelope_shift = self.enveloppe
        inputs = (random_state.rand(n_samples) *
                  (self.inputs_bound[1] - self.inputs_bound[0]) +
                  self.inputs_bound[0])
        # Pattern of the signal
        inputs = sort(inputs)
        pattern = -sin(2 * pi * inputs * self.sine_period)

        # Enveloppe of the signal
        enveloppe = envelope_shift + sin(2 * pi * inputs * envelope_period)

        return (inputs.reshape((n_samples, 1)),
                (pattern * enveloppe).reshape((n_samples, 1)))


def toy_data_quantile(n_samples=50, probs=0.5, pattern=SinePattern(),
                      noise=(1., .2, 0., 1.5), random_state=None):
    """Sine wave toy dataset.

    The target y is computed as a sine curve at of modulated by a sine envelope
    with some period (default 1/3Hz) and mean (default 1).  Moreover, this
    pattern is distorted with a random Gaussian noise with mean 0 and a
    linearly decreasing standard deviation (default from 1.2 at X = 0 to 0.2 at
    X = 1 .  5).

    Parameters
    ----------
    n_samples : int
        Number of samples to generate.

    probs : list or float, shape = [n_quantiles], default=0.5
        Probabilities (quantiles levels).

    pattern : callable, default = SinePattern()
        Callable which generates n_sample 1D data (inputs and targets).

    noise : :rtype: (float, float, float, float)
        Noise parameters (variance, shift, support_min, support_max).

    Returns
    -------
    X : array, shape = [n_samples, 1]
        Input data.

    y : array, shape = [n_sample, 1]
        Targets.

    quantiles : array, shape = [n_samples x n_quantiles]
        True conditional quantiles.
    """
    probs = array(probs, ndmin=1)
    noise_var, noise_shift, noise_min, noise_max = noise

    random_state = check_random_state(random_state)
    pattern.random_state = random_state
    inputs, targets = pattern(n_samples)

    # Noise decreasing std (from noise+0.2 to 0.2)
    noise_std = noise_shift + (noise_var * (noise_max - inputs) /
                               (noise_max - noise_min))
    # Gaussian noise with decreasing std
    add_noise = noise_std * random_state.randn(n_samples, 1)
    observations = targets + add_noise
    quantiles = [targets + array([norm.ppf(p, loc=0., scale=abs(noise_c))
                                  for noise_c in noise_std]) for p in probs]
    return inputs, observations.ravel(), quantiles



"""Simulate supervised and unsupervised."""

from numpy import NaN
from numpy import bool as npbool
from numpy import float as npfloat

from sklearn.utils import check_random_state

# pylint: disable=C0103


def awful(y, p_unsup=.25, p_weaksup=.25, p_weaksup_inner=.25,
          random_state=None):
    """Take a nice dataset and add some NaNs to simulate partially supervised.

    Parameters
    ----------
    y : array, shape = [n_samples, dim_target]
        Targets.

    p_unsup : float, default = .25
        Probability of a single target to be unsupervised.

    p_weaksup : float, default = .25
        Probability of a single target to be weakly supervised.

    p_weaksup_inner : float, default = .25
        Rate of weak supervision.

    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number
        generator; If RandomState instance, random_state is the
        random number generator; If None, the random number
        generator is the RandomState instance used by `np.random`.

    Returns
    -------
    y : array, shape = [n_samples, dim_target]
        Awful targets.
    """
    random_state_internal = check_random_state(random_state)
    awful_targets = y.copy()
    p_sup = 1 - (p_unsup + p_weaksup)
    awful_mask = (random_state_internal
                  .multinomial(1, [p_sup, p_unsup, p_weaksup],
                               size=y.shape[0])
                  .argmax(axis=1))
    awful_targets[awful_mask == 1, :] = NaN
    weaksup_mask = (random_state_internal
                    .binomial(1, 1 - p_weaksup_inner,
                              awful_targets[awful_mask == 2, :].shape)
                    .astype(npfloat))
    weaksup_mask[~weaksup_mask.astype(npbool)] = NaN
    awful_targets[awful_mask == 2, :] = (awful_targets[awful_mask == 2, :] *
                                         weaksup_mask)

    return awful_targets



"""Synthetic datasets for vector-field learning."""

from numpy import arange, sqrt, meshgrid, pi, exp, gradient, floor, hstack

# pylint: disable=C0103


def _gaussian(x_mesh, y_mesh, x_mesh_mean, y_mesh_mean, scale=1):
    """Generate a gaussian.

    Parameters
    ----------
    x_mesh : array, shape = [n_samples, n_samples]
        The inputs mesh x_axis.

    y_mesh : array, shape = [n_samples, n_samples]
        The inputs mesh y_axis.

    x_mesh_mean :
        The x_axis center of the Gaussian.

    y_mesh_mean :
        The y_axis center of the Gaussian.

    scale :
        The scale parameter of the Gaussian. Must be positive.

    Returns
    -------
        field : array, shape = [n_samples, n_samples]
            A scalar field with a Gaussian with some scale centered at
            x_mesh_mean, y_mesh_mean.
    """
    x_mesh_centered = x_mesh - x_mesh_mean
    y_mesh_centered = y_mesh - y_mesh_mean
    return pi ** 2 * exp(- scale / 2 * (x_mesh_centered ** 2 +
                                        y_mesh_centered ** 2)) / sqrt(scale)


def array2mesh(X, side=None):
    """Array to mesh converter.

    Parameters
    ----------
    X : array, shape = [n_samples, 2]
        The inputs array.

    Returns
    -------
    x_mesh : array, shape = [n_samples, n_samples]
        The x_axis of the mesh corresponding to inputs.

    y_mesh : array, shape = [n_samples, n_samples]
        The y_axis of the mesh corresponding to inputs

    """
    if side is None:
        side = int(floor(sqrt(X.shape[0])))
    x_mesh = X[:, 0].reshape((side, side))
    y_mesh = X[:, 1].reshape((side, side))

    return x_mesh, y_mesh


def mesh2array(x_mesh, y_mesh):
    """Mesh to array converter.

    Parameters
    ----------
    x_mesh : array, shape = [n_samples, n_samples]
        The x_axis mesh.

    y_mesh : array, shape = [n_samples, n_samples]
        The y_axis mesh.

    Returns
    -------
    inputs : array, shape = [n_samples, 2]
        The inputs corresponding to the mesh (x_mesh, y_mesh).

    """
    return hstack((x_mesh.ravel().reshape((-1, 1)),
                   y_mesh.ravel().reshape((-1, 1))))


def toy_data_curl_free_mesh(n_samples, loc=25., space=0.5):
    """Curl-Free toy dataset.

    Generate a scalar field as mixture of five gaussians at location:
        - (0   , 0)
        - (0   , loc)
        - ( loc, 0)
        - (-loc, 0)
        - (0   , -loc)
    whith variance equal to 'space'. Then return the gradient of the field.
    The return result is a pair meshes.

    Parameters
    ----------
    n_samples : int
        Number of samples to generate.

    loc: float, optional (default = 25.)
        Centers of the Gaussians.

    space: float, optional (default = .5)
        Variance of the Gaussians.


    Returns
    -------
    X, Y : :rtype: (array, array), shape = [n, n]
            Mesh, X, Y coordinates.

    U, V : :rtype: (array, array), shape = [n, n]
           Mesh, (U, V) velocity at (X, Y) coordinates

    See also
    --------
    operalib.toy_data_curl_free_field
        Generate Curl-Free field.

    operalib.toy_data_div_free_field
        Generate Divergence-Free field.
    """
    x_grid = arange(-1, 1, 2. / sqrt(n_samples))
    y_grid = arange(-1, 1, 2. / sqrt(n_samples))
    x_mesh, y_mesh = meshgrid(x_grid, y_grid)
    field = (_gaussian(x_mesh, y_mesh, -space, 0, loc) +
             _gaussian(x_mesh, y_mesh, space, 0, loc) -
             _gaussian(x_mesh, y_mesh, 0, space, loc) -
             _gaussian(x_mesh, y_mesh, 0, -space, loc))
    v_mesh, u_mesh = gradient(field)

    return (x_mesh, y_mesh), (u_mesh, v_mesh)


def toy_data_div_free_mesh(n_samples, loc=25., space=0.5):
    """Divergence-Free toy dataset.

    Generate a scalar field as mixture of five gaussians at location:
        - (0   , 0)
        - (0   , loc)
        - ( loc, 0)
        - (-loc, 0)
        - (0   , -loc)
    whith variance equal to 'space'. Then return the orthogonal of gradient of
    the field. The return result is a pair of meshes.

    Parameters
    ----------
    n_points : int
        Number of samples to generate.

    loc: float, optional (default = 25.)
        Centers of the Gaussians.

    space: float, optional (default = .5)
        Variance of the Gaussians.


    Returns
    -------
    X, Y : :rtype: (array, array), shape = [n, n]
            Mesh, X, Y coordinates.

    U, V : :rtype: (array, array), shape = [n, n]
           Mesh, (U, V) velocity at (X, Y) coordinates

    See also
    --------
    operalib.toy_data_curl_free_field
        Generate Curl-Free field.

    operalib.toy_data_div_free_field
        Generate Divergence-Free field.
    """
    x_grid = arange(-1, 1, 2. / sqrt(n_samples))
    y_grid = arange(-1, 1, 2. / sqrt(n_samples))
    x_mesh, y_mesh = meshgrid(x_grid, y_grid)
    field = (_gaussian(x_mesh, y_mesh, -space, 0, loc) +
             _gaussian(x_mesh, y_mesh, space, 0, loc) -
             _gaussian(x_mesh, y_mesh, 0, space, loc) -
             _gaussian(x_mesh, y_mesh, 0, -space, loc))
    v_mesh, u_mesh = gradient(field)

    return (x_mesh, y_mesh), (v_mesh, -u_mesh)


def toy_data_curl_free_field(n_samples, loc=25, space=0.5):
    """Curl-Free toy dataset.

    Generate a scalar field as mixture of five gaussians at location:
        - (0   , 0)
        - (0   , loc)
        - ( loc, 0)
        - (-loc, 0)
        - (0   , -loc)
    whith variance equal to 'space'. Then return the gradient of the field.
    The return result is a pair (inputs, targets) of arrays.

    Parameters
    ----------
    n_samples : int
        Number of samples to generate.

    loc: float, optional (default = 25.)
        Centers of the Gaussians.

    space: float, optional (default = .5)
        Variance of the Gaussians.


    Returns
    -------
    X : array, shape = [n_samples, 2]
        Array of evenly space points.

    y : array shape = [n_samples, 2]
        Array corresponding to the velocity at the coordinates present in
        inputs.

    See also
    --------
    operalib.toy_data_curl_free_mesh
        Generate Curl-Free mesh.

    operalib.toy_data_div_free_mesh
        Generate Divergence-Free mesh.
    """
    (x_mesh, y_mesh), (u_mesh, v_mesh) = toy_data_curl_free_mesh(n_samples,
                                                                 loc, space)

    inputs = mesh2array(x_mesh, y_mesh)
    targets = mesh2array(u_mesh, v_mesh)
    return inputs, targets


def toy_data_div_free_field(n_samples, loc=25, space=0.5):
    """Divergence-Free toy dataset.

    Generate a scalar field as mixture of five gaussians at location:
        - (0   , 0)
        - (0   , loc)
        - ( loc, 0)
        - (-loc, 0)
        - (0   , -loc)
    whith variance equal to 'space'. Then return the orthogonal of gradient of
    the field. The return result is a pair (inputs, targets) of arrays.

    Parameters
    ----------
    n_points : int
        Number of samples to generate.

    loc: float, optional (default = 25.)
        Centers of the Gaussians.

    space: float, optional (default = .5)
        Variance of the Gaussians.


    Returns
    -------
    X : array, shape = [n_samples, 2]
        Array of evenly space points.

    y : array shape = [n_samples, 2]
        Array corresponding to the velocity at the coordinates present in
        inputs.

    See also
    --------
    operalib.toy_data_div_free_mesh
        Generate Curl-Free mesh.

    operalib.toy_data_div_free_mesh
        Generate Divergence-Free mesh.
    """
    (x_mesh, y_mesh), (u_mesh, v_mesh) = toy_data_div_free_mesh(n_samples,
                                                                loc, space)

    inputs = mesh2array(x_mesh, y_mesh)
    targets = mesh2array(u_mesh, v_mesh)
    return inputs, targets



"""Simplex coding module."""

from numpy import dot, array, vstack, hstack, ones, zeros, sqrt, asarray

from sklearn.preprocessing import LabelBinarizer
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted

# pylint: disable=W0201,C0103


class SimplexCoding(BaseEstimator, TransformerMixin):
    """Simplex coding."""

    def __init__(self, binarizer=None):
        self.binarizer = binarizer

        # self.simplex_operator_ = None
        # self.binarizer_ = None

    @staticmethod
    def _code_i(dimension):
        """Simplex coding operator (internal).

        https://papers.nips.cc/paper/4764-multiclass-learning-with-simplex-
        coding.pdf
        """
        if dimension > 1:
            block1 = vstack((ones((1, 1)), zeros((dimension - 1, 1))))
            block2 = vstack((-ones((1, dimension)) / dimension,
                             SimplexCoding._code_i(dimension - 1) *
                             sqrt(1. - 1. / (dimension * dimension))))
            return hstack((block1, block2))
        elif dimension == 1:
            return array([1., -1.])
        else:
            raise ValueError('dimension should be at least one.')

    @staticmethod
    def code(dimension):
        """Simplex coding operator."""
        return SimplexCoding._code_i(dimension - 1)

    def fit(self, y):
        """Fit simplex coding

        Parameters
        ----------
        targets : array, shape = [n_samples,] or [n_samples, n_classes]
            Target values. The 2-d array represents the simplex coding for
            multilabel classification.

        Returns
        -------
        self : returns an instance of self.
        """
        if self.binarizer is None:
            self.binarizer_ = LabelBinarizer(neg_label=0, pos_label=1,
                                             sparse_output=True)
        self.binarizer_.fit(y)
        dimension = self.binarizer_.classes_.size
        if dimension > 2:
            self.simplex_operator_ = SimplexCoding.code(dimension)
        else:
            self.simplex_operator_ = ones((1, 1))
        return self

    def transform(self, y):
        """Transform multi-class labels to the simplex code.

        Parameters
        ----------
        targets : array or sparse matrix, shape = [n_samples,] or
                  [n_samples, n_classes]
            Target values. The 2-d matrix represents the simplex code for
            multilabel classification.

        Returns
        -------
        Y : numpy array of shape [n_samples, n_classes - 1]
        """
        check_is_fitted(self, 'simplex_operator_', 'binarizer_')
        dimension = self.binarizer_.classes_.size
        if dimension == 2:
            return self.binarizer_.transform(y).toarray()
        else:
            return self.binarizer_.transform(y).dot(
                asarray(self.simplex_operator_).T)

    def inverse_transform(self, y):
        """Inverse transform."""
        check_is_fitted(self, 'simplex_operator_', 'binarizer_')
        dimension = self.binarizer_.classes_.size
        if dimension == 2:
            return self.binarizer_.inverse_transform(y)
        else:
            return self.binarizer_.inverse_transform(
                dot(y, self.simplex_operator_))



"""
:mod:`operalib.orff` implements Operator-Valued Random Fourier Features
regression.
"""

from scipy.optimize import minimize
from numpy import eye, zeros

from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils import check_X_y, check_array
from sklearn.utils.validation import check_is_fitted
from sklearn.metrics.pairwise import rbf_kernel

from .kernels import DecomposableKernel, RBFCurlFreeKernel
from .risk import ORFFRidgeRisk

# When adding a new kernel, update this table and the _get_kernel_map method
PAIRWISE_KERNEL_FUNCTIONS = {
    'DGauss': DecomposableKernel,
    'DSkewed_chi2': DecomposableKernel,
    'DPeriodic': DecomposableKernel,
    'CurlF': RBFCurlFreeKernel}


class ORFFRidge(BaseEstimator, RegressorMixin):

    def __init__(self, ovkernel='DGauss', loss='LS', lbda=1e-5,
                 A=None, gamma=1., D=1000, skew=0.,
                 solver='L-BFGS-B', solver_params=None):
        self.ovkernel = ovkernel
        self.loss = loss
        self.lbda = lbda
        self.A = A
        self.gamma = gamma
        self.skew = skew
        self.D = D
        self.solver = solver
        self.solver_params = solver_params

    def _validate_params(self):
        # check on self.ovkernel is performed in method __get_kernel
        if self.D < 0:
            raise ValueError('lbda must be a positive integer')
        if self.lbda < 0:
            raise ValueError('lbda must be a positive float')
        # if self.A < 0: # Check whether A is S PD would be really expensive
        #     raise ValueError('A must be a symmetric positive operator')
        if self.gamma is not None:
            if self.gamma < 0:
                raise ValueError('sigma must be positive or default (None)')
        if self.skew is not None:
            if self.skew < 0:
                raise ValueError('sigma must be positive or default (None)')
        # TODO, add supported solver check

    def _get_kernel(self, X, y):
        # When adding a new kernel, update this table and the _get_kernel_map
        # method
        if callable(self.ovkernel):
            ov_kernel = self.ovkernel
        elif isinstance(self.ovkernel, str):
            # 1) check string and assign the right parameters
            if self.ovkernel == 'DGauss':
                self.A_ = self._default_decomposable_op(y)
                kernel_params = {'A': self.A_, 'scalar_kernel': rbf_kernel,
                                 'scalar_kernel_params': {'gamma': self.gamma}}
            elif self.ovkernel == 'DSkewed_chi2':
                self.A_ = self._default_decomposable_op(y)
                kernel_params = {'A': self.A_, 'scalar_kernel': 'skewed_chi2',
                                 'scalar_kernel_params': {'skew': self.skew}}
            elif self.ovkernel == 'CurlF':
                kernel_params = {'gamma': self.gamma}
            else:
                raise NotImplementedError('unsupported kernel')
            # 2) Uses lookup table to select the right kernel from string
            ov_kernel = \
                PAIRWISE_KERNEL_FUNCTIONS[self.ovkernel](**kernel_params)
        else:
            raise NotImplementedError('unsupported kernel')
        return ov_kernel

    def _default_decomposable_op(self, y):
        if self.A is not None:
            return self.A
        elif y.ndim == 2:
            return eye(y.shape[1])
        else:
            return eye(1)

    def _decision_function(self, X):
        pred = self.linop_.get_orff_map(X) * self.coefs_

        return pred.reshape((X.shape[0], self.p)) \
            if self.p > 1 else pred

    def fit(self, X, y):
        """Fit ORFF ridge regression model.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training data.

        y : {array-like}, shape = [n_samples] or [n_samples, n_targets]
            Target values.

        Returns
        -------
        self : returns an instance of self.
        """
        X, y = check_X_y(X, y, ['csr', 'csc', 'coo'],
                         y_numeric=True, multi_output=True)
        self._validate_params()
        self.p = y.shape[1] if y.ndim > 1 else 1

        solver_params = self.solver_params or {}

        self.linop_ = self._get_kernel(X, y)
        self.phix_ = self.linop_.get_orff_map(X, self.D)
        risk = ORFFRidgeRisk(self.lbda, self.loss)
        self.solver_res_ = minimize(risk.functional_grad_val,
                                    zeros(self.phix_.shape[1],
                                          dtype=X.dtype),
                                    args=(y.ravel(), self.phix_, self.linop_),
                                    method=self.solver,
                                    jac=True, options=solver_params)
        self.coefs_ = self.solver_res_.x
        return self

    def predict(self, X):
        """Predict using the ORFF ridge model.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Samples.

        Returns
        -------
        C : {array}, shape = [n_samples] or [n_samples, n_targets]
            Returns predicted values.
        """
        check_is_fitted(self, ['coefs_', 'linop_'], all_or_any=all)
        X = check_array(X)
        return self._decision_function(X)



"""
===================================================
Vector-field Learning with structured output kernel
===================================================

An example to illustrate structured learning with operator-valued kernels.

We compare Operator-valued kernel (OVK) with scikit-learn multi-output ridge
regression.
"""
# Author: Romain Brault <ro.brault@gmail.com>
# License: MIT

# -*- coding: utf-8 -*-
import operalib as ovk

import matplotlib.pyplot as plt

import numpy as np
from numpy.random import RandomState

from sklearn.model_selection import train_test_split
from sklearn.kernel_ridge import KernelRidge


def main():
    """Example of vector-field learning."""

    # Fix a seed
    random_state = RandomState(0)

    # Generate data
    inputs, targets = ovk.toy_data_curl_free_field(n_samples=2000)
    inputs_mesh = ovk.array2mesh(inputs)

    (inputs_train, inputs_test,
     targets_train, targets_test) = train_test_split(inputs, targets,
                                                     train_size=(inputs
                                                                 .shape[0] -
                                                                 40 ** 2),
                                                     random_state=random_state)
    # Add some noise
    targets_train = (targets_train +
                     .175 * random_state.randn(targets_train.shape[0],
                                               targets_train.shape[1]))

    regressor = {'CF':
                 ovk.OVKRidge(ovkernel=ovk.RBFCurlFreeKernel(gamma=2.),
                              lbda=1e-4),
                 'Indep':
                 KernelRidge(kernel='rbf', gamma=.5, alpha=1e-4)}

    # Learning with curl-free
    regressor['CF'].fit(inputs_train, targets_train)
    score_cf = regressor['CF'].score(inputs_test, targets_test)
    print('R2 curl-free ridge: ', score_cf)
    targets_mesh_cf = ovk.array2mesh(regressor['CF'].predict(inputs))

    # Learning with sklearn ridge
    regressor['Indep'].fit(inputs_train, targets_train)
    scode_id = regressor['Indep'].score(inputs_test, targets_test)
    print('R2 independant ridge: ', scode_id)
    targets_mesh_id = ovk.array2mesh(regressor['Indep'].predict(inputs))

    # Plotting
    # pylint: disable=E1101
    fig, axarr = plt.subplots(1, 2, sharex=True, sharey=True, figsize=(14, 7))
    axarr[0].streamplot(inputs_mesh[0], inputs_mesh[1],
                        targets_mesh_cf[0], targets_mesh_cf[1],
                        color=np.sqrt(targets_mesh_cf[0]**2 +
                                      targets_mesh_cf[1]**2),
                        linewidth=.5, cmap=plt.cm.jet, density=2,
                        arrowstyle=u'->')
    axarr[1].streamplot(inputs_mesh[0], inputs_mesh[1],
                        targets_mesh_id[0], targets_mesh_id[1],
                        color=np.sqrt(targets_mesh_id[0]**2 +
                                      targets_mesh_id[1]**2),
                        linewidth=.5, cmap=plt.cm.jet, density=2,
                        arrowstyle=u'->')
    axarr[0].set_ylim([-1, 1])
    axarr[0].set_xlim([-1, 1])
    axarr[0].set_title('Curl-Free Ridge, R2: ' + str(score_cf))
    axarr[1].set_ylim([-1, 1])
    axarr[1].set_xlim([-1, 1])
    axarr[1].set_title('Independant Ridge, R2: ' + str(scode_id))

    fig.suptitle('Vectorfield learning')
    plt.show()

if __name__ == '__main__':
    main()



"""
:mod:`operalib.learningrate` implements the learning rate for (Stochastic)
gradient descent algorithms
"""
# Author: Romain Brault <romain.brault@telecom-paristech.fr> with help from
#         the scikit-learn community.
# License: MIT


class Constant(object):
    r"""Constant learning rate.

    .. math::
        t \mapsto eta

    Attributes
    ----------
    eta : float default 1.
    """

    def __init__(self, eta=1):
        r"""Initialize constant learning rate.

        Parameters
        ----------
        eta : float default 1.
        """
        self.eta = eta

    def __call__(self, step):
        r"""Return learning rate at time t.

        Returns
        -------
        self.eta : float
        """
        return self.eta

    def get_rate(self, step):
        r"""Return learning rate at time t.

        Returns
        -------
        self.eta / t ** self.power : float
        """
        return self.__call__(step)


class InvScaling(object):
    r"""Inverse scaling learnin rate.

    .. math::
        t \mapsto eta0 * t^{-power}

    Parameters
    ----------
    eta0 : float, default 1.

    power : float, default 1.

    Returns
    -------
    InvScaling : Callable
    """

    def __init__(self, eta=1, power=1.):
        r"""Initialize inverse scaling learnin rate.

        Parameters
        ----------
        eta0 : float, default 1.

        power : float, default 1.
        """
        self.eta = eta
        self.power = power

    def __call__(self, step):
        r"""Return learning rate at time t.

        Returns
        -------
        self.eta / t ** self.power : float
        """
        return float(self.eta) / step ** self.power

    def get_rate(self, step):
        r"""Return learning rate at time t.

        Returns
        -------
        self.eta / t ** self.power : float
        """
        return self.__call__(step)



"""OVK learning, unit tests.

The :mod:`sklearn.tests.test_semisuo` tests semisup module.
"""

import operalib as ovk
import numpy as np


def test_semisup_linop():
    """Test ovk.semisup.SemisupLinop."""
    np.random.seed()

    n = 100
    p = 5
    lbda2 = .1

    # supervised indices
    B = np.random.randint(2, size=(n)).astype(np.bool)

    # Graph Laplacian
    n_unsup = np.sum(~B)
    L = np.random.randn(n_unsup, n_unsup)
    L = np.dot(L, L.T)

    U, J = ovk.ridge._SemisupLinop(lbda2, B, L, p).gen()

    y = np.random.randn(n, p)
    # lbda2 * np.dot(L, y[~B, :]).ravel()
    # print(np.concatenate((y[B, :].ravel(),
    #                       lbda2 * np.dot(L, y[~B, :]).ravel())))

    res = np.empty((n, p))
    res[B, :] = y[B, :]
    res[~B] = 0
    assert np.allclose(J * y.ravel(), res.ravel())

    res = np.empty((n, p))
    res[B, :] = y[B, :]
    res[~B] = lbda2 * np.dot(L, y[~B, :])
    assert np.allclose(U * y.ravel(), res.ravel())



"""Synthetic dataset."""


from .vectorfield import (toy_data_curl_free_field,
                          toy_data_div_free_mesh, array2mesh, mesh2array,
                          toy_data_curl_free_mesh, toy_data_div_free_field)
from .quantile import toy_data_quantile
from .awful import awful

__all__ = ['toy_data_curl_free_field', 'toy_data_div_free_mesh',
           'toy_data_div_free_field', 'toy_data_curl_free_mesh',
           'array2mesh', 'mesh2array',
           'toy_data_quantile',
           'awful']



"""
======================================================
Data sparse quantile regression with operator-valued kernels
======================================================

An example to illustrate data sparse quantile regression with operator-valued
kernels.

We compare quantile regression with several levels of data sparsity.
"""

# Author: Maxime Sangnier <maxime.sangnier@gmail.com>
# License: MIT

# -*- coding: utf-8 -*-
import time

import numpy as np
import matplotlib.pyplot as plt

from operalib import Quantile, toy_data_quantile


def main():
    """Example of multiple quantile regression."""

    print("Creating dataset...")
    probs = np.linspace(0.1, 0.9, 5)  # Quantile levels of interest
    x_train, y_train, _ = toy_data_quantile(100, random_state=0)
    x_test, y_test, z_test = toy_data_quantile(1000, probs=probs,
                                               random_state=1)

    print("Fitting...")
    methods = {'Non-sparse QR':
               Quantile(probs=probs, kernel='DGauss', lbda=1e-2, gamma=8.,
                        gamma_quantile=1e-2),
                'Sparse QR':
               Quantile(probs=probs, kernel='DGauss', lbda=1e-2, gamma=8.,
                        gamma_quantile=1e-2, eps=2.5)}
    # Fit on training data
    for name, reg in sorted(methods.items()):
        print(name)
        start = time.time()
        reg.fit(x_train, y_train)

        coefs = reg.model_["coefs"].reshape(y_train.size, probs.size)
        n_sv = np.sum(np.linalg.norm(coefs, axis=1) *
                      reg.lbda / y_train.size > 1e-5)
        print(name + ' leaning time: ', time.time() - start)
        print(name + ' score ', reg.score(x_test, y_test))
        print(name + ' num of support vectors ', n_sv)

    # Plot the estimated conditional quantiles
    plt.figure(figsize=(12, 7))
    for i, method in enumerate(sorted(methods.keys())):
        plt.subplot(2, 2, i + 1)
        plt.plot(x_train, y_train, '.')
        plt.gca().set_prop_cycle(None)
        for quantile in methods[method].predict(x_test):
            plt.plot(x_test, quantile, '-')
        plt.gca().set_prop_cycle(None)
        for prob, quantile in zip(probs, z_test):
            plt.plot(x_test, quantile, '--', label="theoretical {0:0.2f}".format(prob))
        plt.title(method)
        plt.legend(fontsize=8)

        coefs = methods[method].model_["coefs"].reshape(y_train.size,
                                                        probs.size)
        plt.subplot(2, 2, 2 + i + 1)
        plt.plot(np.linalg.norm(coefs, axis=1))
        plt.title("Norm of dual coefs for each point")
    plt.show()

if __name__ == '__main__':
    main()
